# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides interface to G90 alarm panel.

.. note:: Only protocol 1.2 is supported!

The next example queries the device with IP address `10.10.10.250` for the
information - the product name, protocol version, HW versions and such.

>>> import asyncio
>>> from pyg90alarm import G90Alarm
>>> async def main():
       g90 = G90Alarm(host='10.10.10.250')
       s = await g90.host_info
       print(s)
>>> asyncio.run(main())
G90HostInfo(host_guid='<...>',
            product_name='TSV018-C3SIA',
            wifi_protocol_version='1.2',
            cloud_protocol_version='1.1',
            mcu_hw_version='206',
            wifi_hw_version='206',
            gsm_status=3,
            wifi_status=3,
            reserved1=0,
            reserved2=0,
            band_frequency='4264',
            gsm_signal_level=50,
            wifi_signal_level=100)

"""

import logging
from .const import (
    G90Commands, REMOTE_PORT,
    REMOTE_TARGETED_DISCOVERY_PORT,
    LOCAL_TARGETED_DISCOVERY_PORT,
    G90ArmDisarmTypes,
)
from .base_cmd import G90BaseCommand
from .paginated_result import G90PaginatedResult
from .entities.sensor import (G90Sensor, G90SensorTypes)
from .entities.device import G90Device
from .device_notifications import (
    G90DeviceNotifications,
)
from .discovery import G90Discovery
from .targeted_discovery import G90TargetedDiscovery
from .host_info import G90HostInfo
from .host_status import G90HostStatus
from .config import (G90AlertConfig, G90AlertConfigFlags)
from .history import G90History
from .user_data_crc import G90UserDataCRC
from .callback import G90Callback

_LOGGER = logging.getLogger(__name__)


class G90Alarm:  # pylint: disable=too-many-public-methods
    """
    Allows to interact with G90 alarm panel.

    :param str host: Hostname or IP address of the alarm panel. Since the
     protocol is UDP-based it is ok to use broadcast or directed broadcast
     addresses, such as `255.255.255.255` or `10.10.10.255` (the latter assumes
     the device is on `10.10.10.0/24` network)
    :param port: The UDP port of the device where it listens for the
     protocol commands on WiFi interface, currently the devices don't allow it
     to be customized
    :type port: int, optional
    :param sock: The existing socket to operate on, instead of
     creating one internally. Primarily used by the tests to mock the network
     traffic
    :type sock: socket.socket or None, optional
    :param reset_occupancy_interval: The interval upon that the sensors are
     simulated to go into inactive state.
    :type reset_occupancy_interval: int, optional
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, host, port=REMOTE_PORT, sock=None,
                 reset_occupancy_interval=3):
        self._host = host
        self._port = port
        self._sensors = []
        self._devices = []
        self._notifications = None
        self._sock = sock
        self._sensor_cb = None
        self._armdisarm_cb = None
        self._door_open_close_cb = None
        self._reset_occupancy_interval = reset_occupancy_interval
        self._alert_config = None
        self._sms_alert_when_armed = False

    async def command(self, code, data=None):
        """
        Invokes a command against the alarm panel.

        :param code: Command code
        :type code: :class:`.G90Commands`
        :param data: Command data
        :type data: object or None, optional
        :return: :attr:`.G90BaseCommand.result` that contains the result of
         command invocation
        """
        cmd = await G90BaseCommand(
            self._host, self._port, code, data, sock=self._sock).process()
        return cmd.result

    def paginated_result(self, code, start=1, end=None):
        """
        Returns asynchronous generator for a paginated command, that is -
        command operating on a range of records.

        :param code: Command code
        :type code: :class:`.G90Commands`
        :param int start: Starting record position (one-based)
        :param int end: Ending record position (one-based)
        :return: :class:`.G90PaginatedResult` being asynchronous generator
          over the result of command invocation. Each access to the generator
          yields :class:`.G90PaginatedResponse` instance
        """
        return G90PaginatedResult(
            self._host, self._port, code, start, end, sock=self._sock
        ).process()

    @classmethod
    async def discover(cls):
        """
        Initiates discovering devices available in the same network segment, by
        using global broadcast address as the destination.

        :return: List of discovered devices
        :rtype: list[{'guid', 'host', 'port'}]
        """
        return await G90Discovery(
            port=REMOTE_PORT,
            host='255.255.255.255'
        ).process()

    @classmethod
    async def targeted_discover(cls, device_id):
        """
        Initiates discovering devices available in the same network segment
        using targeted protocol, that is - specifying target device GUID in the
        request, so only the specific device should respond to the query.

        :param device_id: GUID of the target device to discover
        :type device_id: str
        :return: List of discovered devices
        :rtype: list[{'guid', 'host', 'port'}]
        """
        return await G90TargetedDiscovery(
            device_id=device_id,
            port=REMOTE_TARGETED_DISCOVERY_PORT,
            local_port=LOCAL_TARGETED_DISCOVERY_PORT,
            host='255.255.255.255'
        ).process()

    @property
    async def sensors(self):
        """
        Provides list of sensors configured in the device. Please note the list
        is cached upon first call, so you need to re-instantiate the class to
        reflect any updates there.

        :return: List of sensors
        :rtype: list(:class:`.G90Sensor`)
        """
        if not self._sensors:
            sensors = self.paginated_result(
                G90Commands.GETSENSORLIST
            )
            async for sensor in sensors:
                obj = G90Sensor(
                    *sensor.data, parent=self, subindex=0,
                    proto_idx=sensor.proto_idx
                )
                self._sensors.append(obj)

            _LOGGER.debug('Total number of sensors: %s', len(self._sensors))

        return self._sensors

    @property
    async def devices(self):
        """
        Provides list of devices (switches) configured in the device. Please
        note the list is cached upon first call, so you need to re-instantiate
        the class to reflect any updates there. Multi-node devices, those
        having multiple ports, are expanded into corresponding number of
        resulting entries.

        :return: List of devices
        :rtype: list(:class:`.G90Device`)
        """
        if not self._devices:
            devices = self.paginated_result(
                G90Commands.GETDEVICELIST
            )
            async for device in devices:
                obj = G90Device(
                    *device.data, parent=self, subindex=0,
                    proto_idx=device.proto_idx
                )
                self._devices.append(obj)
                # Multi-node devices (first node has already been added
                # above
                for node in range(1, obj.node_count):
                    obj = G90Device(
                        *device.data, parent=self,
                        subindex=node, proto_idx=device.proto_idx
                    )
                    self._devices.append(obj)

            _LOGGER.debug('Total number of devices: %s', len(self._devices))

        return self._devices

    @property
    async def host_info(self):
        """
        Provides the device information (for example hardware versions, signal
        levels etc.).

        :return: Device information
        :rtype: Instance of :class:`.G90HostInfo`
        """
        res = await self.command(G90Commands.GETHOSTINFO)
        return G90HostInfo(*res)

    @property
    async def host_status(self):
        """
        Provides the device status (for example, armed or disarmed, configured
        phone number, product name etc.).

        :return: Device information
        :rtype: Instance of :class:`.G90HostStatus`

        """
        res = await self.command(G90Commands.GETHOSTSTATUS)
        return G90HostStatus(*res)

    @property
    async def alert_config(self):
        """
        Retrieves the alert configuration flags from the device. Please note
        the configuration is cached upon first call, so you need to
        re-instantiate the class to reflect any updates there.

        :return: Instance of :class:`.G90AlertConfigFlags` containing the
         alerts configured
        """
        if not self._alert_config:
            self._alert_config = await self._alert_config_uncached()
        return self._alert_config

    async def _alert_config_uncached(self):
        """
        Retrieves the alert configuration flags directly from the device.

        :return: Instance of :class:`.G90AlertConfigFlags` containing the
         alerts configured
        """
        res = await self.command(G90Commands.GETNOTICEFLAG)
        return G90AlertConfig(*res).flags

    async def set_alert_config(self, value):
        """
        It might be possible to implement the async property setter with
        `async_as_sync` decorator, although it might have implications with the
        setter not executed if the program terminates earlier. Hence, for the
        sake of better predictability this is implemented as regular
        (non-property) method
        """
        # Use uncached method retrieving the alert configuration, to ensure the
        # actual value retrieved from the device
        alert_config = await self._alert_config_uncached()
        if alert_config != self._alert_config:
            _LOGGER.warning(
                'Alert configuration changed externally,'
                ' overwriting (read "%s", will be set to "%s")',
                str(alert_config), str(value)
            )
        await self.command(G90Commands.SETNOTICEFLAG, [value])
        # Update the alert configuration stored
        self._alert_config = value

    @property
    async def user_data_crc(self):
        """
        Retieves checksums (CRC) for different on-device databases (history,
        sensors etc.). Might be used to detect if there is a change in a
        particular database.

        .. note:: Note that due to a bug in the firmware CRC for sensos and
          device databases change on each call even if there were no changes

        :return: Instance of :class:`.G90UserDataCRC` containing checksums for
          different databases
        """
        res = await self.command(G90Commands.GETUSERDATACRC)
        return G90UserDataCRC(*res)

    async def history(self, start=1, count=1):
        """
        Retrieves event history from the device.

        :param start: Starting record number (one-based)
        :type start: int
        :param count: Number of records to retrieve
        :type count: int
        :return: List of history entries
        :rtype: list[:class:`.G90History`]
        """
        res = self.paginated_result(G90Commands.GETHISTORY,
                                    start, count)
        history = [G90History(*x.data) async for x in res]
        return history

    async def _internal_sensor_cb(self, idx, name, occupancy=True):
        """
        Callback that invoked both for sensor notifications and door open/close
        alerts, since the logic for both is same and could be reused. Please
        note the callback is for internal use by the class.

        :param int idx: The index of the sensor the callback is invoked for.
         Please note the index is a property of sensor, not the direct index of
         :attr:`sensors` array
        :param str name: The name of the sensor, along with the `idx` parameter
         it is used to look the sensor up from the :attr:`sensors` list
        :param bool occupancy: The flag indicating the target sensor state
         (=occupancy), will always be `True` for callbacks invoked from alarm
         panel notifications, and reflects actual sensor state for device
         alerts (only for `door` type sensors, if door open/close alerts are
         enabled)
        """
        sensors = await self.sensors

        # Fast lookup by direct index
        if idx < len(sensors) and sensors[idx].name == name:
            sensor = [sensors[idx]]
        # Fast lookup failed, perform slow one over the whole sensors list
        else:
            sensor = [
                x for x in sensors
                if x.index == idx and x.name == name
            ]
        if sensor:
            _LOGGER.debug('Found sensor: %s', sensor[0])
            _LOGGER.debug('Setting occupancy to %s (previously %s)',
                          occupancy, sensor[0].occupancy)
            sensor[0].occupancy = occupancy

            # Emulate turning off the occupancy - most of sensors will not
            # notify the device of that, nor the device would emit such
            # notification itself
            def reset_sensor_occupancy(sensor):
                _LOGGER.debug('Resetting occupancy for sensor %s', sensor)
                sensor.occupancy = False
                G90Callback.invoke(sensor.state_callback, sensor.occupancy)

            # Determine if door close notifications are available for the given
            # sensor
            alert_config_flags = await self.alert_config
            door_close_alert_enabled = (
                G90AlertConfigFlags.DOOR_CLOSE in alert_config_flags)
            sensor_is_door = sensor[0].type == G90SensorTypes.DOOR

            # Alarm panel could emit door close alerts (if enabled) for sensors
            # of type `door`, and such event will be used to reset the
            # occupancy for the given sensor. Otherwise, the sensor closing
            # event will be emulated
            if not door_close_alert_enabled or not sensor_is_door:
                _LOGGER.debug("Sensor '%s' is not a door (type %s),"
                              ' or door close alert is disabled'
                              ' (alert config flags %s),'
                              ' closing event will be emulated upon'
                              ' %s seconds',
                              name, sensor[0].type, alert_config_flags,
                              self._reset_occupancy_interval)
                G90Callback.invoke_delayed(
                    self._reset_occupancy_interval,
                    reset_sensor_occupancy, sensor[0])

            # Invoke per-sensor callback if provided
            G90Callback.invoke(sensor[0].state_callback, occupancy)
        else:
            _LOGGER.error('Sensor not found: idx=%s, name=%s', idx, name)

        # Invoke global callback if provided
        G90Callback.invoke(self._sensor_cb, idx, name, occupancy)

    @property
    def sensor_callback(self):
        """
        Get or set sensor activity callback, the callback is invoked when
        sensor activates.

        :type: .. py:function:: ()(idx, name, occupancy)
        """
        return self._sensor_cb

    @sensor_callback.setter
    def sensor_callback(self, value):
        self._sensor_cb = value

    async def _internal_door_open_close_cb(self, idx, name, is_open):
        """
        Callback that invoked when door open/close alert comes from the alarm
        panel. Please note the callback is for internal use by the class.

        .. seealso:: `method`:_internal_sensor_cb for arguments
        """
        # Same internal callback is reused both for door open/close alerts and
        # sensor notifications. The former adds reporting when a door is
        # closed, since the notifications aren't sent for such events
        await self._internal_sensor_cb(idx, name, is_open)
        # Invoke user specified callback if any
        G90Callback.invoke(self._door_open_close_cb, idx, name, is_open)

    @property
    def door_open_close_callback(self):
        """
        Get or set door open/close callback, the callback is invoked when door
        is opened or closed (if corresponding alert is configured on the
        device).

        :type: .. py:function:: ()(idx: int, name: str, is_open: bool)
        """
        return self._door_open_close_cb

    @door_open_close_callback.setter
    def door_open_close_callback(self, value):
        """
        tbd
        """
        self._door_open_close_cb = value

    async def _internal_armdisarm_cb(self, state):
        """
        Callback that invoked when the device is armed or disarmed. Please note
        the callback is for internal use by the class.

        :param state: Device state (armed, disarmed, armed home)
        :type state: :class:`G90ArmDisarmTypes`
        """
        if self._sms_alert_when_armed:
            if state == G90ArmDisarmTypes.DISARM:
                # Disable SMS alerts from the device
                await self.set_alert_config(
                    await self.alert_config & ~G90AlertConfigFlags.SMS_PUSH
                )
            if state in (G90ArmDisarmTypes.ARM_AWAY,
                         G90ArmDisarmTypes.ARM_HOME):
                # Enable SMS alerts from the device
                await self.set_alert_config(
                    await self.alert_config | G90AlertConfigFlags.SMS_PUSH
                )
        G90Callback.invoke(self._armdisarm_cb, state)

    @property
    def armdisarm_callback(self):
        """
        Get or set device arm/disarm callback, the callback is invoked when
        device state changes.

        :type: .. py:function:: ()(state: :class:`.G90ArmDisarmTypes`)
        """
        return self._armdisarm_cb

    @armdisarm_callback.setter
    def armdisarm_callback(self, value):
        self._armdisarm_cb = value

    async def listen_device_notifications(self, sock=None):
        """
        Starts internal listener for device notifications/alerts.

        :param sock: socket instance to listen on, mostly used by tests
        :type: socket.socket
        """
        self._notifications = G90DeviceNotifications(
            sensor_cb=self._internal_sensor_cb,
            door_open_close_cb=self._internal_door_open_close_cb,
            armdisarm_cb=self._internal_armdisarm_cb,
            sock=sock)
        await self._notifications.listen()

    def close_device_notifications(self):
        """
        Closes the listener for device notifications/alerts.
        """
        if self._notifications:
            self._notifications.close()

    async def arm_away(self):
        """
        Arms the device in away mode.
        """
        state = G90ArmDisarmTypes.ARM_AWAY
        await self.command(G90Commands.SETHOSTSTATUS,
                           [state])

    async def arm_home(self):
        """
        Arms the device in home mode.
        """
        state = G90ArmDisarmTypes.ARM_HOME
        await self.command(G90Commands.SETHOSTSTATUS,
                           [state])

    async def disarm(self):
        """
        Disarms the device.
        """
        state = G90ArmDisarmTypes.DISARM
        await self.command(G90Commands.SETHOSTSTATUS,
                           [state])

    @property
    def sms_alert_when_armed(self):
        """
        When enabled, allows to save costs on SMS by having corresponding alert
        enabled only when device is armed.
        """
        return self._sms_alert_when_armed

    @sms_alert_when_armed.setter
    def sms_alert_when_armed(self, value):
        self._sms_alert_when_armed = value



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Extends paginated command for G90 alarm panel providing convenience interface
to work with results of paginated commands.
"""

import logging
from collections import namedtuple
from .paginated_cmd import G90PaginatedCommand
from .const import (
    CMD_PAGE_SIZE,
)

_LOGGER = logging.getLogger(__name__)


class G90PaginatedResponse(
    namedtuple('G90PaginatedResponse', ['proto_idx', 'data'])
):
    """
    Response yielded from the :meth:`.G90PaginatedResult.process` method
    """


class G90PaginatedResult:
    """
    Processes paginated response from G90 corresponding panel commands.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, host, port, code, start=1, end=None, **kwargs):
        # pylint: disable=too-many-arguments
        self._host = host
        self._port = port
        self._code = code
        self._start = start
        self._end = end
        self._kwargs = kwargs

    async def process(self):
        """
        Process paginated response yielding :class:`.G90PaginatedResponse`
        instance for each element.
        """
        page = CMD_PAGE_SIZE
        start = self._start
        count = 0
        while True:
            # The start record number is one-based, so subtract one when
            # calculating the number of the end record for the current
            # iteration
            end = start + page - 1
            # Use the smallest of requested end record number and calculated
            # one (based of page size), allows for number of records less than
            # in page
            if self._end:
                end = min(end, self._end)

            _LOGGER.debug('Invoking paginated command for %s..%s range',
                          start, end)
            cmd = await G90PaginatedCommand(
                host=self._host, port=self._port, code=self._code,
                start=start, end=end,
                **self._kwargs
            ).process()

            # The caller didn't supply the end record number, use the records
            # total since it is now known
            if not self._end:
                self._end = cmd.total
            # The supplied end record number is higher than total records
            # available, reset to the latter
            if self._end > cmd.total:
                _LOGGER.warning('Requested record range (%i) exceeds number of'
                                ' available records (%i), setting to the'
                                ' latter', self._end, cmd.total)
                self._end = cmd.total

            # Produce the resulting records for the consumer
            for idx, data in enumerate(cmd.result):
                # Protocol uses one-based indexes, `start` implies that so no
                # further additions to resulting value is needed.
                # Note the index provided here is running one across multiple
                # pages hence use of `start` variable
                yield G90PaginatedResponse(start + idx, data)

            # Count the number of processed records
            count += cmd.count

            # End the loop if we processed same number of sensors as in the
            # pagination header (or attempted to process more than that by
            # an error)
            _LOGGER.debug('Retrieved %i records in the iteration,'
                          ' %i available in total, target end'
                          ' record number is %i',
                          cmd.count, cmd.total, self._end)
            if cmd.start + cmd.count - 1 >= self._end:
                break
            # Move to the next page for another iteration
            start = start + page

        _LOGGER.debug('Total number of paginated entries:'
                      ' processed %s, expected %s',
                      count,
                      # Again, both end and start record numbers are one-based,
                      # so need to add one to calculate how many records have
                      # been requested
                      self._end - self._start + 1)



import sys
import socket
from unittest.mock import call
import asynctest


PYTHON36 = sys.version_info.major == 3 and sys.version_info.minor == 6


class G90Fixture(asynctest.TestCase):

    def setUp(self):
        def mocked_sendto(data, *_, **__):
            if 'timeout' not in self._testMethodName:
                asynctest.set_read_ready(self.socket_mock, self.loop)
            return len(data)

        self.socket_mock = asynctest.SocketMock()
        self.socket_mock.type = socket.SOCK_DGRAM
        if PYTHON36:
            self.socket_mock.sendto.side_effect = mocked_sendto
        else:
            self.socket_mock.send.side_effect = mocked_sendto

    def assert_callargs_on_sent_data(self, data):
        if PYTHON36:
            call_args = [call(x, None) for x in data]
            self.assertEqual(self.socket_mock.sendto.call_args_list, call_args)
        else:
            call_args = [call(x) for x in data]
            self.assertEqual(self.socket_mock.send.call_args_list, call_args)



import asyncio
import socket
import sys
from unittest.mock import MagicMock
import asynctest
from .fixtures import G90Fixture
sys.path.extend(['src', '../src'])
from pyg90alarm import (   # noqa:E402
    G90Alarm,
)
from pyg90alarm.host_info import (   # noqa:E402
    G90HostInfo, G90HostInfoGsmStatus, G90HostInfoWifiStatus,
)
from pyg90alarm.entities.sensor import (   # noqa:E402
    G90Sensor,
)
from pyg90alarm.entities.device import (   # noqa:E402
    G90Device,
)
from pyg90alarm.history import (   # noqa:E402
    G90History,
)
from pyg90alarm.host_status import (   # noqa:E402
    G90HostStatus,
)
from pyg90alarm.user_data_crc import (   # noqa:E402
    G90UserDataCRC,
)
from pyg90alarm.config import (  # noqa: E402
    G90AlertConfigFlags,
)


class TestG90Alarm(G90Fixture):
    async def test_host_status(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[100,[3,"PHONE","PRODUCT","206","206"]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        res = await g90.host_status
        self.assert_callargs_on_sent_data([b'ISTART[100,100,""]IEND\0'])
        self.assertIsInstance(res, G90HostStatus)
        self.assertEqual(res.host_status, 3)
        self.assertEqual(res.product_name, 'PRODUCT')
        self.assertEqual(res.host_phone_number, 'PHONE')

    async def test_host_info(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[206,' \
            b'["DUMMYGUID","DUMMYPRODUCT",' \
            b'"1.2","1.1","206","206",3,3,0,2,"4242",50,100]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        res = await g90.host_info
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])
        self.assertIsInstance(res, G90HostInfo)
        self.assertEqual(res.host_guid, 'DUMMYGUID')
        self.assertEqual(res.product_name, 'DUMMYPRODUCT')
        self.assertEqual(res.gsm_status, G90HostInfoGsmStatus.OPERATIONAL)
        self.assertEqual(res.wifi_status, G90HostInfoWifiStatus.OPERATIONAL)

    async def test_user_data_crc(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[160,["1","0xab","3","4","5","6"]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        res = await g90.user_data_crc
        self.assert_callargs_on_sent_data([b'ISTART[160,160,""]IEND\0'])
        self.assertIsInstance(res, G90UserDataCRC)
        self.assertEqual(res.sensor_list, '1')
        self.assertEqual(res.device_list, '0xab')
        self.assertEqual(res.history_list, '3')
        self.assertEqual(res.scene_list, '4')
        self.assertEqual(res.ifttt_list, '5')
        self.assertEqual(res.fingerprint_list, '6')

    async def test_devices(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[138,' \
               b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        devices = await g90.devices
        self.assert_callargs_on_sent_data([
            b'ISTART[138,138,[138,[1,10]]]IEND\0',
        ])
        self.assertEqual(len(devices), 1)
        self.assertIsInstance(devices, list)
        self.assertIsInstance(devices[0], G90Device)
        self.assertEqual(devices[0].name, 'Switch')
        self.assertEqual(devices[0].index, 10)

    async def test_multinode_device(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[138,' \
               b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,2,0,""]]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))
        devices = await g90.devices
        self.assert_callargs_on_sent_data([
            b'ISTART[138,138,[138,[1,10]]]IEND\0',
        ])
        self.assertEqual(len(devices), 2)
        self.assertIsInstance(devices, list)
        self.assertIsInstance(devices[0], G90Device)
        self.assertEqual(devices[0].name, 'Switch#1')
        self.assertEqual(devices[0].index, 10)
        self.assertEqual(devices[0].subindex, 0)
        self.assertEqual(devices[1].name, 'Switch#2')
        self.assertEqual(devices[1].index, 10)
        self.assertEqual(devices[1].subindex, 1)

    async def test_control_device(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[138,' \
               b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))
        devices = await g90.devices

        data = b'ISTARTIEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))
        await devices[0].turn_on()
        await devices[0].turn_off()
        self.assert_callargs_on_sent_data([
            b'ISTART[138,138,[138,[1,10]]]IEND\0',
            b'ISTART[137,137,[137,[10,0,0]]]IEND\0',
            b'ISTART[137,137,[137,[10,1,0]]]IEND\0',
        ])

    async def test_single_sensor(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[102,' \
               b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        sensors = await g90.sensors
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
        ])
        self.assertEqual(len(sensors), 1)
        self.assertIsInstance(sensors, list)
        self.assertIsInstance(sensors[0], G90Sensor)
        self.assertEqual(sensors[0].name, 'Remote')
        self.assertEqual(sensors[0].index, 10)

    async def test_multiple_sensors_shorter_than_page(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = [(b'ISTART[102,'
                 b'[[2,1,2],["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
                 ('mocked', 12345)),
                ]

        self.socket_mock.recvfrom.side_effect = data

        sensors = await g90.sensors
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
        ])
        self.assertEqual(len(sensors), 2)
        self.assertIsInstance(sensors, list)
        self.assertIsInstance(sensors[0], G90Sensor)
        self.assertEqual(sensors[0].name, 'Remote 1')
        self.assertEqual(sensors[0].index, 10)
        self.assertIsInstance(sensors[1], G90Sensor)
        self.assertEqual(sensors[1].name, 'Remote 2')
        self.assertEqual(sensors[1].index, 11)

    async def test_multiple_sensors_longer_than_page(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = [(b'ISTART[102,'
                 b'[[11,1,10],'
                 b'["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 3",12,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 4",13,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 5",14,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 6",15,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 7",16,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 8",17,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 9",18,0,10,1,0,32,0,0,16,1,0,""],'
                 b'["Remote 10",19,0,10,1,0,32,0,0,16,1,0,""]'
                 b']]IEND\0',
                 ('mocked', 12345)),
                (b'ISTART[102,'
                 b'[[11,11,1],'
                 b'["Remote 11",20,0,10,1,0,32,0,0,16,1,0,""]'
                 b']]IEND\0',
                 ('mocked', 12345)),
                ]

        self.socket_mock.recvfrom.side_effect = data

        sensors = await g90.sensors
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
            b'ISTART[102,102,[102,[11,11]]]IEND\0',
        ])
        self.assertEqual(len(sensors), 11)
        self.assertIsInstance(sensors, list)
        self.assertIsInstance(sensors[0], G90Sensor)
        self.assertEqual(sensors[0].name, 'Remote 1')
        self.assertEqual(sensors[0].index, 10)

    async def test_sensor_event(self):
        reset_interval = 0.5
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock,
                       reset_occupancy_interval=reset_interval)
        data = [(b'ISTART[102,'
                 b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
                 ('mocked', 12345)),
                (b'ISTART[117,'
                 b'[256]]IEND\0',
                 ('mocked', 12345)),
                ]
        self.socket_mock.recvfrom.side_effect = data

        sensors = await g90.sensors
        future = self.loop.create_future()
        sensor = [x for x in sensors if x.index == 10 and x.name == 'Remote']
        sensor[0].state_callback = lambda *args: future.set_result(True)

        socket_ntfy_mock = asynctest.SocketMock()
        socket_ntfy_mock.type = socket.SOCK_DGRAM
        await g90.listen_device_notifications(socket_ntfy_mock)
        asynctest.set_read_ready(socket_ntfy_mock, self.loop)
        socket_ntfy_mock.recvfrom.return_value = (
            b'[170,[5,[10,"Remote"]]]\0', ('mocked', 12345))
        await asyncio.wait([future], timeout=0.1)
        g90.close_device_notifications()
        # Once passed this implies the G90Alarm sensor callback works as
        # expected, as it updates the occupancy states of the sensor
        self.assertEqual(sensor[0].occupancy, True)

        # Re-create future since the callback will be invoked again when the
        # occupancy state is reset
        future = self.loop.create_future()
        # Verify the occupancy state is reset upon configured interval
        await asyncio.sleep(reset_interval)
        self.assertEqual(sensor[0].occupancy, False)

    async def test_armdisarm_callback(self):
        future = self.loop.create_future()
        armdisarm_cb = MagicMock()
        armdisarm_cb.side_effect = lambda *args: future.set_result(True)
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        g90.armdisarm_callback = armdisarm_cb
        socket_ntfy_mock = asynctest.SocketMock()
        socket_ntfy_mock.type = socket.SOCK_DGRAM
        await g90.listen_device_notifications(socket_ntfy_mock)
        asynctest.set_read_ready(socket_ntfy_mock, self.loop)
        socket_ntfy_mock.recvfrom.return_value = (
            b'[170,[1,[1]]]\0', ('mocked', 12345))
        await asyncio.wait([future], timeout=0.1)
        g90.close_device_notifications()
        armdisarm_cb.assert_called_once_with(1)

    async def test_door_open_close_callback(self):
        future = self.loop.create_future()
        door_open_close_cb = MagicMock()
        door_open_close_cb.side_effect = lambda *args: future.set_result(True)
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = [(b'ISTART[102,'
                 b'[[1,1,1],["Door",100,0,1,1,0,32,0,0,16,1,0,""]]]IEND\0',
                 ('mocked', 12345)),
                (b'ISTART[117,'
                 b'[256]]IEND\0',
                 ('mocked', 12345)),
                # The GETNOTICEFLAG command is invoked twice, for each of
                # alerts simulated below
                (b'ISTART[117,'
                 b'[256]]IEND\0',
                 ('mocked', 12345)),
                ]
        self.socket_mock.recvfrom.side_effect = data

        g90.door_open_close_callback = door_open_close_cb
        socket_ntfy_mock = asynctest.SocketMock()
        socket_ntfy_mock.type = socket.SOCK_DGRAM
        await g90.listen_device_notifications(socket_ntfy_mock)
        # Simulate two device alerts - for opening and then closing the door
        socket_ntfy_mock.recvfrom.side_effect = [
            (b'[208,[4,100,1,1,"Door","DUMMYGUID",1631545189,0,[""]]]\0',
             ('mocked', 12345)),
            (b'[208,[4,100,1,0,"Door","DUMMYGUID",1631545189,0,[""]]]\0',
             ('mocked', 12345)),
        ]
        # Signal the first alert is ready
        asynctest.set_read_ready(socket_ntfy_mock, self.loop)
        await asyncio.wait([future], timeout=0.1)
        # Corresponding sensor should turn to occupied (=door opened)
        sensors = await g90.sensors
        self.assertEqual(sensors[0].occupancy, True)

        # Signal the second alert is ready, the future has to be re-created as
        # the corresponding callback will be fired again
        future = self.loop.create_future()
        asynctest.set_read_ready(socket_ntfy_mock, self.loop)
        await asyncio.wait([future], timeout=0.1)
        # The sensor should become inactive (=door closed)
        sensors = await g90.sensors
        self.assertEqual(sensors[0].occupancy, False)

        g90.close_device_notifications()

    async def test_arm_away(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTARTIEND\0', ('mocked', 12345))
        await g90.arm_away()
        self.assert_callargs_on_sent_data([
            b'ISTART[101,101,[101,[1]]]IEND\0'
        ])

    async def test_arm_home(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTARTIEND\0', ('mocked', 12345))
        await g90.arm_home()
        self.assert_callargs_on_sent_data([
            b'ISTART[101,101,[101,[2]]]IEND\0'
        ])

    async def test_disarm(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTARTIEND\0', ('mocked', 12345))
        await g90.disarm()
        self.assert_callargs_on_sent_data([
            b'ISTART[101,101,[101,[3]]]IEND\0'
        ])

    async def test_history(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[200,[[50,1,5],'
            b'[3,33,7,254,"Sensor 1",1630147285,""],'
            b'[2,3,0,0,"",1630142877,""],'
            b'[2,5,0,0,"",1630142871,""],'
            b'[2,4,0,0,"",1630142757,""],'
            b'[3,100,126,1,"Sensor 2",1630142297,""]]]IEND\0',
            ('mocked', 12345))
        history = await g90.history(count=5)
        self.assertEqual(len(history), 5)
        self.assertIsInstance(history[0], G90History)
        self.assert_callargs_on_sent_data([
            b'ISTART[200,200,[200,[1,5]]]IEND\0'
        ])

    async def test_alert_config(self):
        """ Tests for retrieving alert configuration from the device """
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        data = b'ISTART[117,[1]]IEND\0'
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        res = await g90.alert_config
        self.assert_callargs_on_sent_data([b'ISTART[117,117,""]IEND\0'])
        self.assertIsInstance(res, G90AlertConfigFlags)

    async def test_set_alert_config(self):
        """ Tests for setting alert configuration to the the device """
        g90 = G90Alarm(host="mocked", port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.side_effect = [
            (b"ISTART[117,[1]]IEND\0", ("mocked", 12345)),
            (b"ISTART[117,[3]]IEND\0", ("mocked", 12345)),
            (b"ISTARTIEND\0", ("mocked", 12345)),
        ]

        await g90.set_alert_config(
            await g90.alert_config
            | G90AlertConfigFlags.AC_POWER_FAILURE  # noqa:W503
            | G90AlertConfigFlags.HOST_LOW_VOLTAGE  # noqa:W503
        )
        self.assert_callargs_on_sent_data(
            [
                b'ISTART[117,117,""]IEND\0',
                b'ISTART[117,117,""]IEND\0',
                b"ISTART[116,116,[116,[9]]]IEND\0",
            ]
        )
        # Validate we retrieve same alert configuration just has been set
        self.assertEqual(
            await g90.alert_config,
            G90AlertConfigFlags.AC_POWER_FAILURE
            | G90AlertConfigFlags.HOST_LOW_VOLTAGE  # noqa:W503
        )

    async def test_sms_alert_when_armed(self):
        """ Tests for enabling SMS alerts when device is armed """
        future = self.loop.create_future()
        armdisarm_cb = MagicMock()
        armdisarm_cb.side_effect = lambda *args: future.set_result(True)
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        g90.armdisarm_callback = armdisarm_cb
        g90.sms_alert_when_armed = True
        socket_ntfy_mock = asynctest.SocketMock()
        socket_ntfy_mock.type = socket.SOCK_DGRAM
        await g90.listen_device_notifications(socket_ntfy_mock)
        asynctest.set_read_ready(socket_ntfy_mock, self.loop)
        self.socket_mock.recvfrom.side_effect = [
            # First command to get alert configuration is from
            # `G90Alarm.alert_config` property
            (b"ISTART[117,[1]]IEND\0", ("mocked", 12345)),
            # Second command for same is invoked by `G90Alarm.set_alert_config`
            # that checks if alert config has been modified externally
            (b"ISTART[117,[1]]IEND\0", ("mocked", 12345)),
            (b"ISTARTIEND\0", ("mocked", 12345)),
        ]
        socket_ntfy_mock.recvfrom.side_effect = [
            (b'[170,[1,[1]]]\0', ('mocked', 12345)),
        ]
        await asyncio.wait([future], timeout=0.1)
        g90.close_device_notifications()
        self.assert_callargs_on_sent_data(
            [
                b'ISTART[117,117,""]IEND\0',
                b'ISTART[117,117,""]IEND\0',
                b"ISTART[116,116,[116,[513]]]IEND\0",
            ]
        )

    async def test_sms_alert_when_disarmed(self):
        """ Tests for disabling SMS alerts when device is disarmed """
        future = self.loop.create_future()
        armdisarm_cb = MagicMock()
        armdisarm_cb.side_effect = lambda *args: future.set_result(True)
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        g90.armdisarm_callback = armdisarm_cb
        g90.sms_alert_when_armed = True
        socket_ntfy_mock = asynctest.SocketMock()
        socket_ntfy_mock.type = socket.SOCK_DGRAM
        await g90.listen_device_notifications(socket_ntfy_mock)
        asynctest.set_read_ready(socket_ntfy_mock, self.loop)
        self.socket_mock.recvfrom.side_effect = [
            # See above for the clarification on command sequence
            (b"ISTART[117,[513]]IEND\0", ("mocked", 12345)),
            (b"ISTART[117,[513]]IEND\0", ("mocked", 12345)),
            (b"ISTARTIEND\0", ("mocked", 12345)),
        ]
        socket_ntfy_mock.recvfrom.side_effect = [
            (b'[170,[1,[3]]]\0', ('mocked', 12345)),
        ]
        await asyncio.wait([future], timeout=0.1)
        g90.close_device_notifications()
        self.assert_callargs_on_sent_data(
            [
                b'ISTART[117,117,""]IEND\0',
                b'ISTART[117,117,""]IEND\0',
                b"ISTART[116,116,[116,[1]]]IEND\0",
            ]
        )

    async def test_sensor_disable(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.side_effect = [
            (
                b'ISTART[102,'
                b'[[2,1,2],'
                b'["Night Light1",11,0,138,0,0,33,0,0,17,1,0,""],'
                b'["Night Light2",10,0,138,0,0,33,0,0,17,1,0,""]'
                b']]IEND\0',
                ('mocked', 12345)
            ),
            (
                b'ISTART[102,'
                b'[[2,2,1],'
                b'["Night Light2",10,0,138,0,0,33,0,0,17,1,0,""]'
                b']]IEND\0',
                ('mocked', 12345)
            ),
            (b"ISTARTIEND\0", ("mocked", 12345)),
        ]

        sensors = await g90.sensors
        self.assertEqual(sensors[1].enabled, True)
        await sensors[1].set_enabled(False)
        self.assertEqual(sensors[1].enabled, False)
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
            b'ISTART[102,102,[102,[2,2]]]IEND\0',
            b'ISTART[103,103,[103,'
            b'["Night Light2",10,0,138,0,0,32,0,0,17,1,0,2,"060A0600"]'
            b']]IEND\0',
        ])

    async def test_sensor_disable_externally_modified(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.side_effect = [
            (
                b'ISTART[102,'
                b'[[2,1,2],'
                b'["Night Light1",11,0,138,0,0,33,0,0,17,1,0,""],'
                b'["Night Light2",10,0,138,0,0,33,0,0,17,1,0,""]'
                b']]IEND\0',
                ('mocked', 12345)
            ),
            (
                b'ISTART[102,'
                b'[[2,2,1],'
                b'["Night Light2",10,0,138,0,0,1,0,0,17,1,0,""]'
                b']]IEND\0',
                ('mocked', 12345)
            ),
            (b"ISTARTIEND\0", ("mocked", 12345)),
        ]

        sensors = await g90.sensors
        self.assertEqual(sensors[1].enabled, True)
        await sensors[1].set_enabled(False)
        self.assertEqual(sensors[1].enabled, True)
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
            b'ISTART[102,102,[102,[2,2]]]IEND\0',
        ])

    async def test_sensor_unsupported_disable(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.side_effect = [
            (
                b'ISTART[102,'
                b'[[1,1,1],["Water",10,0,7,0,0,33,0,0,17,1,0,""]'
                b']]IEND\0',
                ('mocked', 12345)
            ),
            (b"ISTARTIEND\0", ("mocked", 12345)),
        ]

        sensors = await g90.sensors
        self.assertEqual(sensors[0].enabled, True)
        await sensors[0].set_enabled(False)
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,10]]]IEND\0',
        ])

    async def test_device_unsupported_disable(self):
        g90 = G90Alarm(host='mocked', port=12345, sock=self.socket_mock)
        self.socket_mock.recvfrom.side_effect = [
            (
                b'ISTART[138,'
                b'[[1,1,1],["Water",10,0,7,0,0,33,0,0,17,1,0,""]'
                b']]IEND\0',
                ('mocked', 12345)
            ),
            (b"ISTARTIEND\0", ("mocked", 12345)),
        ]

        devices = await g90.devices
        self.assertEqual(devices[0].enabled, True)
        await devices[0].set_enabled(False)
        self.assert_callargs_on_sent_data([
            b'ISTART[138,138,[138,[1,10]]]IEND\0',
        ])



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Discovers G90 alarm panels.
"""

import asyncio
import logging

from .base_cmd import G90BaseCommand
from .host_info import G90HostInfo
from .const import G90Commands

_LOGGER = logging.getLogger(__name__)


class G90DiscoveryProtocol:
    """
    tbd

    :meta private:
    """
    def __init__(self, parent):
        """
        tbd
        """
        self._parent = parent

    def connection_made(self, transport):
        """
        tbd
        """

    def connection_lost(self, exc):
        """
        tbd
        """

    def datagram_received(self, data, addr):
        """
        tbd
        """
        try:
            ret = self._parent.from_wire(data)
            host_info = G90HostInfo(*ret)
            _LOGGER.debug('Received from %s:%s: %s', addr[0], addr[1], ret)
            res = {
                'guid': host_info.host_guid,
                'host': addr[0],
                'port': addr[1]
            }
            res.update(host_info._asdict())
            _LOGGER.debug('Discovered device: %s', res)
            self._parent.add_device(res)

        except Exception as exc:  # pylint: disable=broad-except
            _LOGGER.warning('Got exception, ignoring: %s', exc)

    def error_received(self, exc):
        """
        tbd
        """


class G90Discovery(G90BaseCommand):
    """
    tbd
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, timeout=10, **kwargs):
        """
        tbd
        """
        # pylint: disable=too-many-arguments
        super().__init__(code=G90Commands.GETHOSTINFO, timeout=timeout,
                         **kwargs)
        self._discovered_devices = []

    async def process(self):
        """
        tbd
        """
        _LOGGER.debug('Attempting device discovery...')
        transport, _ = await self._create_connection()
        transport.sendto(self.to_wire())
        await asyncio.sleep(self._timeout)
        transport.close()
        _LOGGER.debug('Discovered %s devices', len(self.devices))
        return self.devices

    @property
    def devices(self):
        """
        tbd
        """
        return self._discovered_devices

    def add_device(self, value):
        """
        tbd
        """
        self._discovered_devices.append(value)

    def _proto_factory(self):
        """
        tbd
        """
        return G90DiscoveryProtocol(self)



import unittest

from .test_base_commands import *  # noqa: F401,F403
from .test_paginated_commands import *  # noqa: F401,F403
from .test_discovery import *  # noqa: F401,F403
from .test_notifications import *  # noqa: F401,F403
from .test_alarm import *  # noqa: F401, F403
from .test_callback import *  # noqa: F401, F403

if __name__ == '__main__':
    unittest.main(verbosity=3)



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements callbacks.
"""

import asyncio
from functools import (partial, wraps)
import logging

_LOGGER = logging.getLogger(__name__)


class G90Callback:
    """
    tbd
    """
    @staticmethod
    def invoke(callback, *args, **kwargs):
        """
        tbd
        """
        if not callback:
            return

        _LOGGER.debug('Attempting to invoke callback %s'
                      ' (args: %s, kwargs: %s)',
                      callback, args, kwargs)

        if not asyncio.iscoroutinefunction(callback):
            def async_wrapper(func):
                """
                Wraps the regular callback function into coroutine, so it could
                later be created as async task.
                """
                @wraps(func)
                async def wrapper(*args, **kwds):
                    return func(*args, **kwds)
                return wrapper

            callback = async_wrapper(callback)

        if hasattr(asyncio, 'create_task'):
            task = asyncio.create_task(callback(*args, **kwargs))
        else:
            # Python 3.6 has only `ensure_future` method
            task = asyncio.ensure_future(callback(*args, **kwargs))

        def reap_callback_exception(task):
            """
            Reaps an exception (if any) from the task logging it, to prevent
            `asyncio` reporting that task exception was never retrieved.
            """
            exc = task.exception()
            if exc:
                _LOGGER.error('Got exception when invoking'
                              " callback '%s(...)':",
                              task.get_coro().__qualname__,
                              exc_info=exc, stack_info=False)

        task.add_done_callback(reap_callback_exception)

    @staticmethod
    def invoke_delayed(delay, callback, *args, **kwargs):
        """
        tbd
        """
        if hasattr(asyncio, 'get_running_loop'):
            loop = asyncio.get_running_loop()
        else:
            # Python 3.6 has no `get_running_loop`, only `get_event_loop`
            loop = asyncio.get_event_loop()
        loop.call_later(delay, partial(callback, *args, **kwargs))


def async_as_sync(func):
    """
    Invokes an async function as regular one via :py:func:`G90Callback.invoke`.
    One of possible use cases is implementing property setter for async code,
    where the function could be used an decorator:

    .. code-block:: python

     @property
     async def a_property(...):
         ...

     @a_property.setter
     @async_as_sync
     async def a_property(...):
         ...

    Since the function internally submits the wrapped async code as
    :py:mod:`asyncio` task, it execution isn't guaranteed as the program could
    be terminated earlier that it is processed in the loop.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        return G90Callback.invoke(func, *args, **kwargs)
    return wrapper



import sys
from .fixtures import G90Fixture
sys.path.extend(['src', '../src'])

from pyg90alarm.base_cmd import (  # noqa:E402
    G90BaseCommand
)
from pyg90alarm.exceptions import (G90Error, G90TimeoutError)  # noqa:E402


class TestG90BaseCommand(G90Fixture):
    async def test_network_unreachable(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.side_effect = OSError('Host unreachable')

        with self.assertRaises(OSError) as cm:
            await g90.process()
        self.assertIn("Host unreachable", cm.exception.args)

    async def test_wrong_host(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTARTIEND\0', ('another_host', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn('Received response from wrong host another_host,'
                      ' expected from mocked', cm.exception.args)

    async def test_wrong_port(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTARTIEND\0', ('mocked', 54321))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn('Received response from wrong port 54321,'
                      ' expected from 12345', cm.exception.args)

    async def test_timeout(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206,
            timeout=0.1, retries=2, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (b'', ('mocked', 12345))

        with self.assertRaises(G90TimeoutError):
            await g90.process()
        self.assert_callargs_on_sent_data([
            b'ISTART[206,206,""]IEND\0',
            b'ISTART[206,206,""]IEND\0',
        ])

    async def test_wrong_format(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[IEND\0', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn("Unable to parse response as JSON: '['",
                      cm.exception.args)
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])

    async def test_no_response(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (b'', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn('Missing start marker in data', cm.exception.args)
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])

    async def test_empty_response(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTARTIEND\0', ('mocked', 12345))

        await g90.process()
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])

    async def test_no_code_response(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[]IEND\0', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn("Missing code in response: '[]'", cm.exception.args)
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])

    async def test_wrong_code_response(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[106,[""]]IEND\0', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn('Wrong response - received code 106, expected code 206',
                      cm.exception.args)
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])

    async def test_no_data_response(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[206]IEND\0', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn("Missing data in response: '[206]'",
                      cm.exception.args)
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])

    async def test_no_end_marker(self):
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=206, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[206,[]]IEND', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn('Missing end marker in data', cm.exception.args)
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Represents various configuration aspects of the alarm panel.
"""
from enum import IntFlag
from collections import namedtuple


class G90AlertConfigFlags(IntFlag):
    """ Alert configuration flags, used bitwise """
    AC_POWER_FAILURE = 1
    AC_POWER_RECOVER = 2
    ARM_DISARM = 4
    HOST_LOW_VOLTAGE = 8
    SENSOR_LOW_VOLTAGE = 16
    WIFI_AVAILABLE = 32
    WIFI_UNAVAILABLE = 64
    DOOR_OPEN = 128
    DOOR_CLOSE = 256
    SMS_PUSH = 512
    UNKNOWN1 = 2048
    UNKNOWN2 = 8192


class G90AlertConfig(namedtuple('G90AlertConfig', ['flags_data'])):
    """
    Represents alert configuration as received from the alarm panel.
    """

    @property
    def flags(self):
        """
        :return: Instance of :class:`G90AlertConfigFlags` that provides
         symbolic names to corresponding flag bits
        """
        return G90AlertConfigFlags(self.flags_data)



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Discovers G90 alarm panel devices with specific ID.
"""

import logging
from collections import namedtuple
from .discovery import G90Discovery
from .exceptions import G90Error

_LOGGER = logging.getLogger(__name__)

INCOMING_FIELDS = [
    'message',
    'product_name',
    'wifi_protocol_version',
    'cloud_protocol_version',
    'mcu_hw_version',
    'fw_version',
    'gsm_status',
    'wifi_status',
    'server_status',
    'reserved1',
    'reserved2',
    'gsm_signal_level',
    'wifi_signal_level'
]


class G90TargetedDiscoveryInfo(namedtuple('G90TargetedDiscoveryInfo',
                                          INCOMING_FIELDS)):
    """
    tbd

    :meta private:
    """


class G90TargetedDiscoveryProtocol:
    """
    tbd

    :meta private:
    """
    def __init__(self, device_id, parent):
        """
        tbd
        """
        self._parent = parent
        self._device_id = device_id

    def connection_made(self, transport):
        """
        tbd
        """

    def connection_lost(self, exc):
        """
        tbd
        """

    def datagram_received(self, data, addr):
        """
        tbd
        """
        try:
            _LOGGER.debug('Received from %s:%s: %s', addr[0], addr[1], data)
            decoded = data.decode('utf-8', errors='ignore')
            if not decoded.endswith('\0'):
                raise G90Error('Invalid discovery response')
            host_info = G90TargetedDiscoveryInfo(*decoded[:-1].split(','))
            if host_info.message != 'IWTAC_PROBE_DEVICE_ACK':
                raise G90Error('Invalid discovery response')
            res = {'guid': self._device_id,
                   'host': addr[0],
                   'port': addr[1]}
            res.update(host_info._asdict())
            _LOGGER.debug('Discovered device: %s', res)
            self._parent.add_device(res)
        except Exception as exc:  # pylint: disable=broad-except
            _LOGGER.warning('Got exception, ignoring: %s', exc)

    def error_received(self, exc):
        """
        tbd
        """


class G90TargetedDiscovery(G90Discovery):
    """
    tbd
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, device_id, **kwargs):
        """
        tbd
        """

        super().__init__(**kwargs)
        self._device_id = device_id

    def to_wire(self):  # pylint: disable=no-self-use
        """
        tbd
        """
        return bytes(f'IWTAC_PROBE_DEVICE,{self._device_id}\0', 'ascii')

    def _proto_factory(self):
        """
        tbd
        """
        return G90TargetedDiscoveryProtocol(self._device_id,
                                            self)



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides support for basic commands of G90 alarm panel.
"""

import logging
import json
import asyncio
from typing import NamedTuple, Optional
from .exceptions import (G90Error, G90TimeoutError)


_LOGGER = logging.getLogger(__name__)


class G90Header(NamedTuple):
    """
    Represents JSON structure of the header used in alarm panel commands.
    Note that typing.NamedTuple is used (instead of collections.namedtuple) to
    provide support for Python 3.6 and higher, while providing default values.

    :meta private:
    """
    code: Optional[int] = None
    data: Optional[str] = None


class G90DeviceProtocol:
    """
    tbd

    :meta private:
    """
    def __init__(self):
        """
        tbd
        """
        self._data = None

    @property
    def future_data(self):
        """
        tbd
        """
        return self._data

    @future_data.setter
    def future_data(self, value):
        """
        tbd
        """
        self._data = value

    def connection_made(self, transport):
        """
        tbd
        """

    def connection_lost(self, exc):
        """
        tbd
        """

    def datagram_received(self, data, addr):
        """
        tbd
        """
        if asyncio.isfuture(self._data):
            if self._data.done():
                _LOGGER.warning('Excessive packet received'
                                ' from %s:%s: %s',
                                addr[0], addr[1], data)
                return
            self._data.set_result((*addr, data))

    def error_received(self, exc):
        """
        tbd
        """
        if asyncio.isfuture(self._data) and not self._data.done():
            self._data.set_exception(exc)


class G90BaseCommand:
    """
    tbd
    """
    # pylint: disable=too-many-instance-attributes
    # Lock need to be shared across all of the class instances
    _sk_lock = asyncio.Lock()

    def __init__(self, host, port, code,
                 data=None, local_port=None,
                 timeout=3.0, retries=3, sock=None):
        """
        tbd
        """
        # pylint: disable=too-many-arguments
        self._remote_host = host
        self._remote_port = port
        self._local_port = local_port
        self._code = code
        self._timeout = timeout
        self._retries = retries
        self._data = '""'
        self._result = None
        if data:
            self._data = json.dumps([code, data],
                                    # No newlines to be inserted
                                    indent=None,
                                    # No whitespace around entities
                                    separators=(',', ':'))
        self._resp = G90Header()
        self._sock = sock

    def _proto_factory(self):  # pylint: disable=no-self-use
        """
        tbd
        """
        return G90DeviceProtocol()

    async def _create_connection(self):
        """
        tbd
        """
        try:
            loop = asyncio.get_running_loop()
        except AttributeError:
            loop = asyncio.get_event_loop()

        if self._sock:
            _LOGGER.debug('Using provided socket %s', self._sock)
            transport, protocol = await loop.create_datagram_endpoint(
                self._proto_factory,
                sock=self._sock)
        else:
            _LOGGER.debug('Creating UDP endpoint for %s:%s',
                          self.host, self.port)
            extra_kwargs = {}
            if self._local_port:
                extra_kwargs['local_addr'] = ('0.0.0.0', self._local_port)

            transport, protocol = await loop.create_datagram_endpoint(
                self._proto_factory,
                remote_addr=(self.host, self.port),
                **extra_kwargs,
                allow_broadcast=True)

        return transport, protocol

    def to_wire(self):
        """
        tbd
        """
        wire = bytes(f'ISTART[{self._code},{self._code},{self._data}]IEND\0',
                     'utf-8')
        _LOGGER.debug('Encoded to wire format %s', wire)
        return wire

    def from_wire(self, data):
        """
        tbd
        """
        _LOGGER.debug('To be decoded from wire format %s', data)
        self._parse(data.decode('utf-8', errors='ignore'))
        return self._resp.data

    def _parse(self, data):
        """
        tbd
        """
        if not data.startswith('ISTART'):
            raise G90Error('Missing start marker in data')
        if not data.endswith('IEND\0'):
            raise G90Error('Missing end marker in data')
        payload = data[6:-5]
        _LOGGER.debug("Decoded from wire: JSON string '%s'", payload)

        resp = None
        if payload:
            try:
                resp = json.loads(payload)
            except json.JSONDecodeError as exc:
                raise G90Error('Unable to parse response as JSON:'
                               f" '{payload}'") from exc

            if not isinstance(resp, list):
                raise G90Error('Mailformed response,'
                               f" 'list' expected: '{payload}'")

        if resp is not None:
            self._resp = G90Header(*resp)
            _LOGGER.debug('Parsed from wire: %s', self._resp)

            if not self._resp.code:
                raise G90Error(f"Missing code in response: '{payload}'")
            # Check there is data if the response is non-empty
            if not self._resp.data:
                raise G90Error(f"Missing data in response: '{payload}'")

            if self._resp.code != self._code:
                raise G90Error(
                    'Wrong response - received code '
                    f"{self._resp.code}, expected code {self._code}")

    @property
    def result(self):
        """
        tbd
        """
        return self._result

    @property
    def host(self):
        """
        tbd
        """
        return self._remote_host

    @property
    def port(self):
        """
        tbd
        """
        return self._remote_port

    async def process(self):
        """
        tbd
        """

        transport, protocol = await self._create_connection()
        attempts = self._retries
        while True:
            attempts = attempts - 1
            try:
                loop = asyncio.get_running_loop()
            except AttributeError:
                loop = asyncio.get_event_loop()
            protocol.future_data = loop.create_future()
            async with self._sk_lock:
                _LOGGER.debug('(code %s) Sending request to %s:%s',
                              self._code, self.host, self.port)
                transport.sendto(self.to_wire())
                done, _ = await asyncio.wait([protocol.future_data],
                                             timeout=self._timeout)
            if protocol.future_data in done:
                break
            # Cancel the future to signal protocol handler it is no longer
            # valid, the future will be re-created on next retry
            protocol.future_data.cancel()
            if not attempts:
                transport.close()
                raise G90TimeoutError()
            _LOGGER.debug('Timed out, retrying')
        transport.close()
        (host, port, data) = protocol.future_data.result()
        _LOGGER.debug('Received response from %s:%s', host, port)
        if self.host != '255.255.255.255':
            if self.host != host or host == '255.255.255.255':
                raise G90Error(f'Received response from wrong host {host},'
                               f' expected from {self.host}')
        if self.port != port:
            raise G90Error(f'Received response from wrong port {port},'
                           f' expected from {self.port}')

        ret = self.from_wire(data)
        self._result = ret
        return self

    def __repr__(self):
        """
        tbd
        """
        return f'Command: {self._code}, request: {self._data},' \
            f' response: {self._resp.data}'



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements support for notifications/alerts sent by G90 alarm panel.
"""

import json
import logging
from collections import namedtuple
import asyncio
from .callback import G90Callback
from .const import (
    G90MessageTypes,
    G90NotificationTypes,
    G90AlertTypes,
    G90AlertStateChangeTypes,
    G90ArmDisarmTypes,
    G90AlertSources,
)

_LOGGER = logging.getLogger(__name__)


class G90Message(namedtuple('G90Message',
                            ['code', 'data'])):
    """
    tbd

    :meta private:
    """


class G90Notification(namedtuple('G90Notification',
                                 ['kind', 'data'])):
    """
    tbd

    :meta private:
    """


class G90ZoneInfo(namedtuple('G90ZoneInfo',
                             ['idx', 'name'])):
    """
    tbd

    :meta private:
    """


class G90ArmDisarmInfo(namedtuple('G90ArmDisarmInfo',
                                  ['state'])):
    """
    tbd

    :meta private:
    """


class G90DeviceAlert(namedtuple('G90DeviceAlert',
                                ['type', 'event_id', 'source', 'state',
                                 'zone_name', 'device_id', 'unix_time',
                                 'resv4', 'other'])):
    """
    tbd

    :meta private:
    """


class G90DeviceNotificationProtocol:
    """
    tbd

    :meta private:
    """
    def __init__(self, armdisarm_cb=None, sensor_cb=None,
                 door_open_close_cb=None):
        """
        tbd
        """
        self._armdisarm_cb = armdisarm_cb
        self._sensor_cb = sensor_cb
        self._door_open_close_cb = door_open_close_cb

    def connection_made(self, transport):
        """
        tbd
        """

    def connection_lost(self, exc):
        """
        tbd
        """

    def _handle_notification(self, addr, notification):
        # Sensor activity notification
        if notification.kind == G90NotificationTypes.SENSOR_ACTIVITY:
            g90_zone_info = G90ZoneInfo(*notification.data)
            _LOGGER.debug('Sensor notification: %s', g90_zone_info)
            G90Callback.invoke(self._sensor_cb,
                               g90_zone_info.idx,
                               g90_zone_info.name)
            return

        # Arm/disarm notification
        if notification.kind == G90NotificationTypes.ARM_DISARM:
            g90_armdisarm_info = G90ArmDisarmInfo(
                *notification.data)
            # Map the state received from the device to corresponding enum
            state = G90ArmDisarmTypes(g90_armdisarm_info.state)
            _LOGGER.debug('Arm/disarm notification: %s',
                          state)
            G90Callback.invoke(self._armdisarm_cb,
                               state)
            return

        _LOGGER.warning('Unknown notification received from %s:%s:'
                        ' kind %s, data %s',
                        addr[0], addr[1], notification.kind, notification.data)

    def _handle_alert(self, addr, alert):
        if alert.type == G90AlertTypes.DOOR_OPEN_CLOSE:
            is_open = (
                alert.source == G90AlertSources.SENSOR and alert.state == 1
            ) or alert.source == G90AlertSources.DOORBELL
            _LOGGER.debug('Door open_close alert: %s', alert)
            G90Callback.invoke(self._door_open_close_cb,
                               alert.event_id, alert.zone_name,
                               is_open)
            return

        if alert.type == G90AlertTypes.STATE_CHANGE:
            # Define the mapping between device state received in the alert, to
            # common `G90ArmDisarmTypes` enum that is used when setting device
            # arm state and received in the corresponding notifications. The
            # primary reason is to unify state as passed down to the callbacks.
            # The map covers only subset of state changes pertinent to
            # arm/disarm state changes
            alarm_arm_disarm_state_map = {
                G90AlertStateChangeTypes.ARM_HOME: G90ArmDisarmTypes.ARM_HOME,
                G90AlertStateChangeTypes.ARM_AWAY: G90ArmDisarmTypes.ARM_AWAY,
                G90AlertStateChangeTypes.DISARM: G90ArmDisarmTypes.DISARM
            }

            state = alarm_arm_disarm_state_map.get(alert.event_id, None)
            if state:
                # We received the device state change related to arm/disarm,
                # invoke the corresponding callback
                _LOGGER.debug('Arm/disarm state change: %s', state)
                G90Callback.invoke(self._armdisarm_cb, state)
                return

        _LOGGER.warning('Unknown alert received from %s:%s:'
                        ' type %s, data %s',
                        addr[0], addr[1], alert.type, alert)

    def datagram_received(self, data, addr):  # pylint:disable=R0911
        """
        tbd
        """
        s_data = data.decode('utf-8')
        if not s_data.endswith('\0'):
            _LOGGER.error('Missing end marker in data')
            return
        payload = s_data[:-1]
        _LOGGER.debug('Received device message from %s:%s: %s',
                      addr[0], addr[1], payload)
        try:
            message = json.loads(payload)
            g90_message = G90Message(*message)
        except json.JSONDecodeError as exc:
            _LOGGER.error("Unable to parse device message '%s' as JSON: %s",
                          payload, exc)
            return
        except TypeError as exc:
            _LOGGER.error("Device message '%s' is malformed: %s",
                          payload, exc)
            return

        # Device notifications
        if g90_message.code == G90MessageTypes.NOTIFICATION:
            try:
                data = G90Notification(*g90_message.data)
            except TypeError as exc:
                _LOGGER.error('Bad notification received from %s:%s: %s',
                              addr[0], addr[1], exc)
                return
            self._handle_notification(addr, data)
            return

        # Device alerts
        if g90_message.code == G90MessageTypes.ALERT:
            try:
                data = G90DeviceAlert(*g90_message.data)
            except TypeError as exc:
                _LOGGER.error('Bad alert received from %s:%s: %s',
                              addr[0], addr[1], exc)
                return
            self._handle_alert(addr, data)
            return

        _LOGGER.warning('Unknown message received from %s:%s: %s',
                        addr[0], addr[1], message)


class G90DeviceNotifications:
    """
    tbd
    """
    def __init__(self, port=12901, armdisarm_cb=None, sensor_cb=None,
                 door_open_close_cb=None, sock=None):
        # pylint: disable=too-many-arguments
        self._notification_transport = None
        self._port = port
        self._armdisarm_cb = armdisarm_cb
        self._sensor_cb = sensor_cb
        self._door_open_close_cb = door_open_close_cb
        self._sock = sock

    def proto_factory(self):
        """
        tbd
        """
        return G90DeviceNotificationProtocol(
            self._armdisarm_cb, self._sensor_cb, self._door_open_close_cb)

    async def listen(self):
        """
        tbd
        """
        try:
            loop = asyncio.get_running_loop()
        except AttributeError:
            loop = asyncio.get_event_loop()

        if self._sock:
            _LOGGER.debug('Using provided socket %s', self._sock)
            (self._notification_transport,
             _protocol) = await loop.create_datagram_endpoint(
                self.proto_factory,
                sock=self._sock)
        else:
            _LOGGER.debug('Creating UDP endpoint for 0.0.0.0:%s',
                          self._port)
            (self._notification_transport,
             _protocol) = await loop.create_datagram_endpoint(
                self.proto_factory,
                local_addr=('0.0.0.0', self._port))

    def close(self):
        """
        tbd
        """
        if self._notification_transport:
            self._notification_transport.close()



import sys
from .fixtures import G90Fixture
sys.path.extend(['src', '../src'])

from pyg90alarm.paginated_cmd import (   # noqa:E402
    G90PaginatedCommand,
)
from pyg90alarm.exceptions import G90Error  # noqa:E402


class TestG90PaginatedCommand(G90Fixture):

    async def test_missing_pagination(self):
        g90 = G90PaginatedCommand(
            host='mocked', port=12345, code=102, start=1, end=1,
            sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[102,[[]]]IEND\0', ('mocked', 12345))

        with self.assertRaises(G90Error, ) as cm:
            await g90.process()
        self.assertIn(cm.exception.args[0],
                      ['Wrong pagination data [] -'
                       ' <lambda>() missing 3 required positional arguments:'
                       " 'total', 'start', and 'count'",
                       'Wrong pagination data [] -'
                       ' __new__() missing 3 required positional arguments:'
                       " 'total', 'start', and 'count'"])

        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,1]]]IEND\0'
        ])

    async def test_no_paginated_data(self):
        g90 = G90PaginatedCommand(
            host='mocked', port=12345, code=102, start=1, end=1,
            sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[102,[[1,1,1]]]IEND\0', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn('Truncated data provided in paginated response -'
                      ' expected 1 entities as per response, received 0',
                      cm.exception.args)
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,1]]]IEND\0'
        ])

    async def test_partial_paginated_data(self):
        g90 = G90PaginatedCommand(
            host='mocked', port=12345, code=102, start=1, end=2,
            sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[102,[[1,1,1],[""]]]IEND\0', ('mocked', 12345))

        await g90.process()
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,2]]]IEND\0'
        ])

    async def test_extra_paginated_data(self):
        g90 = G90PaginatedCommand(
            host='mocked', port=12345, code=102, start=1, end=1,
            sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (
            b'ISTART[102,[[2,1,2],[""],[""]]]IEND\0', ('mocked', 12345))

        with self.assertRaises(G90Error) as cm:
            await g90.process()
        self.assertIn('Extra data provided in paginated response -'
                      ' expected 1 entities as per request, received 2',
                      cm.exception.args)
        self.assert_callargs_on_sent_data([
            b'ISTART[102,102,[102,[1,1]]]IEND\0'
        ])



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
History protocol entity.
"""

import time
from collections import namedtuple

INCOMING_FIELDS = [
    'log_type',  # (1 or 3 - alarm, 2 or 4 - notification)
    'param1',    # (type 1: 1 - SOS, 2 - tamper alarm; type 3 - device ID; type
                 #  2 - 5 stayarm, 3 - disarm, 4 - awayarm )
    'param2',    # (type 3: device type)
    'param3',
    'sensor_name',
    'unix_time',
    'rest',
]


class G90History(namedtuple('G90History', INCOMING_FIELDS)):
    """
    tbd
    """

    @property
    def datetime(self):
        """
        tbd
        """
        return time.ctime(self.unix_time)

    def __repr__(self):
        """
        tbd
        """
        return super().__repr__() + f'(datetime={str(self.datetime)})'



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides interface to sensors of G90 alarm panel.
"""

import logging
from collections import namedtuple
from enum import IntEnum, IntFlag
from ..definitions.sensors import SENSOR_DEFINITIONS
from ..const import G90Commands

# Common protocol fields across read and write operations
COMMON_FIELDS = [
    'parent_name',
    'index',
    'room_id',
    'type_id',
    'subtype',
    'timeout',
    'user_flag_data',
    'baudrate',
    'protocol_id',
    'reserved_data',
    'node_count',
]

# Incoming (read operation) protocol fields
INCOMING_FIELDS = COMMON_FIELDS + [
    'mask',
    'private_data',
]

# Outgoing (write operation) protocol fields
OUTGOING_FIELDS = COMMON_FIELDS + [
    'rx',
    'tx',
    'private_data',
]


class G90SensorReservedFlags(IntFlag):
    """
    Reserved flags of the sensor.

    :meta private:
    """
    NONE = 0
    CAN_READ = 16
    CAN_READ_EXT = 32
    CAN_WRITE = 1


class G90SensorUserFlags(IntFlag):
    """
    User flags of the sensor.

    :meta private:
    """
    NONE = 0
    ENABLED = 1
    ARM_DELAY = 2
    DETECT_DOOR = 4
    DOOR_CHIME = 8
    INDEPENDENT_ZONE = 16
    ALERT_WHEN_AWAY_AND_HOME = 32
    ALERT_WHEN_AWAY = 64
    SUPPORTS_UPDATING_SUBTYPE = 512     # Only relevant for cord sensors


class G90SensorProtocols(IntEnum):
    """
    Protocol types for the sensors.

    :meta private:
    """
    RF_1527 = 0
    RF_2262 = 1
    RF_PRIVATE = 2
    RF_SLIDER = 3
    CORD = 5
    WIFI = 4
    USB = 6


class G90SensorTypes(IntEnum):
    """
    Sensor types.

    :meta private:
    """
    DOOR = 1
    GLASS = 2
    GAS = 3
    SMOKE = 4
    SOS = 5
    VIB = 6
    WATER = 7
    INFRARED = 8
    IN_BEAM = 9
    REMOTE = 10
    RFID = 11
    DOORBELL = 12
    BUTTONID = 13
    WATCH = 14
    FINGER_LOCK = 15
    SUBHOST = 16
    REMOTE_2_4G = 17
    CORD_SENSOR = 126
    SOCKET = 128
    SIREN = 129
    CURTAIN = 130
    SLIDINGWIN = 131
    AIRCON = 136
    TV = 137
    NIGHTLIGHT = 138
    SOCKET_2_4G = 140
    SIREN_2_4G = 141
    SWITCH_2_4G = 142
    TOUCH_SWITCH_2_4G = 143
    CURTAIN_2_4G = 144
    CORD_DEV = 254
    UNKNOWN = 255


_LOGGER = logging.getLogger(__name__)


class G90Sensor:  # pylint:disable=too-many-instance-attributes
    """
    Interacts with sensor on G90 alarm panel.

    :param args: Pass-through positional arguments for for interpreting
     protocol fields
    :param parent: Instance of :class:`.G90Alarm` the sensor is associated
     with
    :type parent: :class:`.G90Alarm`
    :param int subindex: Index of the sensor within multi-channel devices
     (those having multiple nodes)
    :param int proto_idx: Index of the sensor within list of sensors as
     retrieved from the alarm panel
    :param kwargs: Pass-through keyword arguments for for interpreting protocol
     fields
    """
    def __init__(self, *args, parent, subindex, proto_idx, **kwargs):
        self._protocol_incoming_data_kls = (
            namedtuple('G90SensorIncomingData', INCOMING_FIELDS)
        )
        self._protocol_outgoing_data_kls = (
            namedtuple('G90SensorOutgoingData', OUTGOING_FIELDS)
        )
        self._protocol_data = self._protocol_incoming_data_kls(*args, **kwargs)
        self._parent = parent
        self._subindex = subindex
        self._occupancy = False
        self._state_callback = None
        self._proto_idx = proto_idx

        self._definition = None
        # Get sensor definition corresponds to the sensor type/subtype if any
        for s_def in SENSOR_DEFINITIONS:
            if (
                s_def.type == self._protocol_data.type_id
                and s_def.subtype == self._protocol_data.subtype  # noqa:W503
            ):
                self._definition = s_def
                break

    @property
    def name(self):
        """
        Returns sensor name, accounting for multi-channel entities (single
        protocol entity results in multiple :class:`.G90Sensor` instances).

        :return: Sensor name
        :rtype: str
        """
        if self._protocol_data.node_count == 1:
            return self._protocol_data.parent_name
        return f'{self._protocol_data.parent_name}#{self._subindex + 1}'

    @property
    def state_callback(self):
        """
        Returns state callback the sensor might have set.

        :return: Sensor state callback
        :rtype: object
        """
        return self._state_callback

    @state_callback.setter
    def state_callback(self, value):
        """
        Sets callback for the state changes of the sensor.

        :param object value: Sensor state callback
        """
        self._state_callback = value

    @property
    def occupancy(self):
        """
        Occupancy (occupied/not occupied, or triggered/not triggered)
        for the sensor.

        :return: Sensor occupancy
        :rtype: bool
        """
        return self._occupancy

    @occupancy.setter
    def occupancy(self, value):
        """
        Sets occupancy state for the sensor.

        :param bool value: Sensor occupancy
        """
        self._occupancy = value

    @property
    def protocol(self):
        """
        Returns protocol type of the sensor.

        :return: Protocol type
        :rtype: int
        """
        return G90SensorProtocols(self._protocol_data.protocol_id)

    @property
    def type(self):
        """
        Returns type of the sensor.

        :return: Sensor type
        :rtype: int
        """
        return G90SensorTypes(self._protocol_data.type_id)

    @property
    def reserved(self):
        """
        Returns reserved flags (read/write mode) for the sensor.

        :return: Reserved flags
        :rtype: int
        """
        return G90SensorReservedFlags(self._protocol_data.reserved_data)

    @property
    def user_flag(self):
        """
        Returns user flags for the sensor (disabled/enabled, arming type etc).

        :return: User flags
        :rtype: int
        """
        return G90SensorUserFlags(self._protocol_data.user_flag_data)

    @property
    def node_count(self):
        """
        Returns number of nodes (channels) for the sensor.

        :return: Number of nodes
        :rtype: int
        """
        return self._protocol_data.node_count

    @property
    def parent(self):
        """
        Returns parent :class:`.G90Alarm` instance the sensor is associated
        with.

        :return: Parent :class:`.G90Alarm` instance
        :rtype: :class:`.G90Alarm`
        """
        return self._parent

    @property
    def index(self):
        """
        Returns index (internal position) of the sensor in the G90 alarm panel.

        :return: Internal sensor position
        :rtype: int
        """
        return self._protocol_data.index

    @property
    def subindex(self):
        """
        Returns index of the sensor within multi-node device.

        :return: Index of sensor in multi-node device.
        :rtype: int
        """
        return self._subindex

    @property
    def supports_enable_disable(self):
        """
        Indicates if disabling/enabling the sensor is supported.

        :return: Support for enabling/disabling the sensor
        :rtype: bool
        """
        return self._definition is not None

    @property
    def enabled(self):
        """
        Indicates if the sensor is enabled.

        :return: If sensor is enabled
        :rtype: bool
        """
        return self.user_flag & G90SensorUserFlags.ENABLED

    async def set_enabled(self, value):
        """
        Sets disabled/enabled state of the sensor.

        :param bool value: Whether to enable or disable the sensor
        """
        if not self.supports_enable_disable:
            _LOGGER.warning(
                'Manipulating with enable/disable for sensor index=%s'
                ' is unsupported - no sensor definition for'
                ' type=%s, subtype=%s',
                self.index,
                self._protocol_data.type_id,
                self._protocol_data.subtype
            )

            return

        # Refresh actual sensor data from the alarm panel before modifying it.
        # This implies the sensor is at the same position within sensor list
        # (`_proto_index`) as it has been read initially from the alarm panel
        # when instantiated.
        _LOGGER.debug(
            'Refreshing sensor at index=%s, position in protocol list=%s',
            self.index, self._proto_idx
        )
        sensors = self.parent.paginated_result(
            G90Commands.GETSENSORLIST,
            start=self._proto_idx, end=self._proto_idx
        )

        # Compare actual sensor data from what the sensor has been instantiated
        # from, and abort the operation if out-of-band changes are detected.
        _sensor_pos, sensor_data = [x async for x in sensors][0]
        if self._protocol_incoming_data_kls(
            *sensor_data
        ) != self._protocol_data:
            _LOGGER.error(
                "Sensor index=%s '%s' has been changed externally,"
                " refusing to alter its enable/disable state",
                self.index,
                self.name
            )
            return

        # Modify the value of the user flag setting enabled/disabled one
        # appropriately.
        user_flag = self.user_flag
        if value:
            user_flag |= G90SensorUserFlags.ENABLED
        else:
            user_flag &= ~G90SensorUserFlags.ENABLED

        # Re-instantiate the protocol data with modified user flags
        _data = self._protocol_data._asdict()
        _data['user_flag_data'] = user_flag
        self._protocol_data = self._protocol_incoming_data_kls(**_data)

        _LOGGER.debug(
            'Sensor index=%s: %s enabled flag, resulting user_flag %s',
            self._protocol_data.index,
            'Setting' if value else 'Clearing',
            self.user_flag
        )

        # Generate protocol data from write operation, deriving values either
        # from fields read from the sensor, or from the sensor definition - not
        # all fields are present during read, only in definition.
        outgoing_data = self._protocol_outgoing_data_kls(
            parent_name=self._protocol_data.parent_name,
            index=self._protocol_data.index,
            room_id=self._protocol_data.room_id,
            type_id=self._protocol_data.type_id,
            subtype=self._protocol_data.subtype,
            timeout=self._protocol_data.timeout,
            user_flag_data=self._protocol_data.user_flag_data,
            baudrate=self._protocol_data.baudrate,
            protocol_id=self._protocol_data.protocol_id,
            reserved_data=self._definition.reserved_data,
            node_count=self._protocol_data.node_count,
            rx=self._definition.rx,
            tx=self._definition.tx,
            private_data=self._definition.private_data,
        )
        # Modify the sensor
        await self._parent.command(
            G90Commands.SETSINGLESENSOR, list(outgoing_data)
        )

    def __repr__(self):
        """
        Returns string representation of the sensor.

        :return: String representation
        :rtype: str
        """
        return super().__repr__() + f'(name={str(self.name)}' \
            f' type={str(self.type)}' \
            f' protocol={str(self.protocol)}' \
            f' occupancy={self.occupancy}' \
            f' user flag={str(self.user_flag)}' \
            f' reserved={str(self.reserved)})'



import sys
import asynctest
from .fixtures import G90Fixture
sys.path.extend(['src', '../src'])
from pyg90alarm.discovery import (   # noqa:E402
    G90Discovery,
)
from pyg90alarm.targeted_discovery import (   # noqa:E402
    G90TargetedDiscovery,
)
from pyg90alarm.const import (  # noqa:E402
    REMOTE_PORT, REMOTE_TARGETED_DISCOVERY_PORT,
    LOCAL_TARGETED_DISCOVERY_PORT,
)


class TestG90Discovery(G90Fixture):
    async def test_discovery(self):
        discovery_data = [
            (b'ISTART[206,["DUMMYGUID1","","","","","",0,0,0,0,"",0,0]]IEND\0',
             ('mocked1', 12345)),
            (b'ISTART[206,["DUMMYGUID2","","","","","",0,0,0,0,"",0,0]]IEND\0',
             ('mocked2', 54321)),
        ]

        def discovery_recvfrom(*_args, **_kwargs):
            try:
                ret = discovery_data.pop(0)
            except IndexError:
                ret = None
            # Indicate the socket is ready for another read if there is mocked
            # data avaialble still
            if discovery_data:
                asynctest.set_read_ready(self.socket_mock, self.loop)
            return ret

        g90 = G90Discovery(host='255.255.255.255',
                           port=REMOTE_PORT,
                           timeout=0.1, sock=self.socket_mock)
        self.socket_mock.recvfrom.side_effect = discovery_recvfrom
        discovered = await g90.process()
        self.assertEqual(discovered[0]['guid'], 'DUMMYGUID1')
        self.assertEqual(discovered[0]['host'], 'mocked1')
        self.assertEqual(discovered[0]['port'], 12345)
        self.assertEqual(discovered[1]['guid'], 'DUMMYGUID2')
        self.assertEqual(discovered[1]['host'], 'mocked2')
        self.assertEqual(discovered[1]['port'], 54321)
        self.assert_callargs_on_sent_data([b'ISTART[206,206,""]IEND\0'])


class TestG90TargetedDiscovery(G90Fixture):
    async def test_targeted_discovery(self):
        data = b'IWTAC_PROBE_DEVICE_ACK,TSV018-3SIA' \
               b',1.2,1.1,206,1.8,3,3,1,0,2,50,100\0'

        g90 = G90TargetedDiscovery(
            device_id='DUMMYGUID',
            host='255.255.255.255',
            port=REMOTE_TARGETED_DISCOVERY_PORT,
            local_port=LOCAL_TARGETED_DISCOVERY_PORT,
            timeout=0.1, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))
        discovered = await g90.process()
        self.assertEqual(discovered[0]['guid'], 'DUMMYGUID')
        self.assertEqual(discovered[0]['host'], 'mocked')
        self.assertEqual(discovered[0]['port'], 12345)
        self.assert_callargs_on_sent_data([b'IWTAC_PROBE_DEVICE,DUMMYGUID\0'])

    async def test_targeted_discovery_wrong_response_start_marker(self):
        data = b'IWTAC_PROBE_DEVICE_ACK_BAD,TSV018-3SIA' \
               b',1.2,1.1,206,1.8,3,3,1,0,2,50,100\0'

        g90 = G90TargetedDiscovery(
            device_id='DUMMYGUID',
            host='255.255.255.255',
            port=REMOTE_TARGETED_DISCOVERY_PORT,
            local_port=LOCAL_TARGETED_DISCOVERY_PORT,
            timeout=0.1, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        with self.assertLogs(level='WARNING') as cm:
            await g90.process()
            self.assertEqual(cm.output, [
                'WARNING:pyg90alarm.targeted_discovery:'
                'Got exception, ignoring: Invalid discovery response'
            ])

    async def test_targeted_discovery_wrong_response_end_marker(self):
        data = b'IWTAC_PROBE_DEVICE_ACK,TSV018-3SIA' \
               b',1.2,1.1,206,1.8,3,3,1,0,2,50,100'

        g90 = G90TargetedDiscovery(
            device_id='DUMMYGUID',
            host='255.255.255.255',
            port=REMOTE_TARGETED_DISCOVERY_PORT,
            local_port=LOCAL_TARGETED_DISCOVERY_PORT,
            timeout=0.1, sock=self.socket_mock)
        self.socket_mock.recvfrom.return_value = (data, ('mocked', 12345))

        with self.assertLogs(level='WARNING') as cm:
            await g90.process()
            self.assertEqual(cm.output, [
                'WARNING:pyg90alarm.targeted_discovery:'
                'Got exception, ignoring: Invalid discovery response'
            ])



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.rst').read_text(encoding='utf-8')

setup(
    name='pyg90alarm',
    setup_requires=['setuptools_scm'],
    use_scm_version={
        "local_scheme": "no-local-version"
    },
    description='G90 Alarm system protocol',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/hostcc/pyg90alarm',
    author='Ilia Sotnikov',
    author_email='hostcc@gmail.com',

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Home Automation',
        'Topic :: System :: Hardware',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='g90, alarm, protocol',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.6, <4',
    install_requires=[],

    extras_require={
        'dev': [
            'check-manifest',
        ],
        'test': [
            'coverage',
            'asynctest',
        ],
        'docs': [
            'sphinx',
            'sphinx-rtd-theme',
        ],
    },

    project_urls={
        'Bug Reports': 'https://github.com/hostcc/pyg90alarm/issues',
        'Source': 'https://github.com/hostcc/pyg90alarm/',
    },
)



import asyncio
import sys
from .fixtures import G90Fixture
from pyg90alarm.callback import (   # noqa:E402
    async_as_sync,
)
sys.path.extend(['src', '../src'])


class TestG90Callback(G90Fixture):
    async def test_async_as_sync_decorator(self):
        """ Tests for `async_as_sync` decorator """
        class TestProperty:
            _prop = None

            @property
            async def test_property(self):
                return self._prop

            @test_property.setter
            @async_as_sync
            async def test_property(self, value):
                self._prop = value

        test_instance = TestProperty()
        test_instance.test_property = 'test value'
        # Allow the async task the decorator submits to finish
        await asyncio.sleep(0)
        self.assertEqual(await test_instance.test_property, 'test value')



import sys
import asyncio
from unittest.mock import MagicMock
import asynctest
from .fixtures import G90Fixture
sys.path.extend(['src', '../src'])
from pyg90alarm.device_notifications import (   # noqa:E402
    G90DeviceNotifications,
)


class TestG90Notifications(G90Fixture):
    async def test_device_notification_missing_header(self):
        def sock_data_awaitable(*args):
            future.set_result(True)
            return b'[170]\0', ('mocked', 12345)

        future = self.loop.create_future()
        notifications = G90DeviceNotifications(sock=self.socket_mock)
        await notifications.listen()
        self.socket_mock.recvfrom.side_effect = sock_data_awaitable
        asynctest.set_read_ready(self.socket_mock, self.loop)
        with self.assertLogs(level='ERROR') as cm:
            await asyncio.wait([future], timeout=0.1)
            self.assertIn(
                cm.output[0], [
                    'ERROR:pyg90alarm.device_notifications:Device message'
                    " '[170]' is malformed: <lambda>() missing 1 required"
                    " positional argument: 'data'",
                    'ERROR:pyg90alarm.device_notifications:Device message'
                    " '[170]' is malformed: __new__() missing 1 required"
                    " positional argument: 'data'",
                ]
            )
        notifications.close()

    async def test_device_notification_malformed_message(self):
        def sock_data_awaitable(*args):
            future.set_result(True)
            return b'[170,[1,[1]]\0', ('mocked', 12345)

        future = self.loop.create_future()
        notifications = G90DeviceNotifications(sock=self.socket_mock)
        await notifications.listen()
        self.socket_mock.recvfrom.side_effect = sock_data_awaitable
        asynctest.set_read_ready(self.socket_mock, self.loop)
        with self.assertLogs(level='ERROR') as cm:
            await asyncio.wait([future], timeout=0.1)
            self.assertIn(
                'ERROR:pyg90alarm.device_notifications:'
                "Unable to parse device message '[170,[1,[1]]' as JSON:",
                cm.output[0]
            )
        notifications.close()

    async def test_device_notification_missing_end_marker(self):
        def sock_data_awaitable(*args):
            future.set_result(True)
            return b'[170,[1,[1]]]', ('mocked', 12345)

        future = self.loop.create_future()
        notifications = G90DeviceNotifications(sock=self.socket_mock)
        await notifications.listen()
        self.socket_mock.recvfrom.side_effect = sock_data_awaitable
        asynctest.set_read_ready(self.socket_mock, self.loop)
        with self.assertLogs(level='ERROR') as cm:
            await asyncio.wait([future], timeout=0.1)
            self.assertEqual(
                cm.output[0],
                'ERROR:pyg90alarm.device_notifications:'
                'Missing end marker in data'
            )
        notifications.close()

    async def test_wrong_device_notification_format(self):
        def sock_data_awaitable(*args):
            future.set_result(True)
            return b'[170,[1]]\0', ('mocked', 12345)

        future = self.loop.create_future()
        notifications = G90DeviceNotifications(sock=self.socket_mock)
        await notifications.listen()
        self.socket_mock.recvfrom.side_effect = sock_data_awaitable
        asynctest.set_read_ready(self.socket_mock, self.loop)
        with self.assertLogs(level='ERROR') as cm:
            await asyncio.wait([future], timeout=0.1)
            self.assertIn(cm.output[0], [
                'ERROR:pyg90alarm.device_notifications:'
                'Bad notification received from mocked:12345:'
                " <lambda>() missing 1 required positional argument: 'data'",
                'ERROR:pyg90alarm.device_notifications:'
                'Bad notification received from mocked:12345:'
                " __new__() missing 1 required positional argument: 'data'",
            ])
        notifications.close()

    async def test_wrong_device_alert_format(self):
        def sock_data_awaitable(*args):
            future.set_result(True)
            return (b'[208,[]]\0', ('mocked', 12345))
        future = self.loop.create_future()
        notifications = G90DeviceNotifications(sock=self.socket_mock)
        await notifications.listen()
        self.socket_mock.recvfrom.side_effect = sock_data_awaitable
        asynctest.set_read_ready(self.socket_mock, self.loop)
        with self.assertLogs(level='ERROR') as cm:
            await asyncio.wait([future], timeout=0.1)
            self.assertIn(cm.output[0], [
                'ERROR:pyg90alarm.device_notifications:'
                'Bad alert received from mocked:12345:'
                " <lambda>() missing 9 required positional arguments: 'type',"
                " 'event_id', 'source', 'state', 'zone_name', 'device_id',"
                " 'unix_time', 'resv4', and 'other'",
                'ERROR:pyg90alarm.device_notifications:'
                'Bad alert received from mocked:12345:'
                " __new__() missing 9 required positional arguments: 'type',"
                " 'event_id', 'source', 'state', 'zone_name', 'device_id',"
                " 'unix_time', 'resv4', and 'other'",
            ])
        notifications.close()

    async def test_unknown_device_notification(self):
        def sock_data_awaitable(*args):
            future.set_result(True)
            return b'[170,[999,[1]]]\0', ('mocked', 12345)

        future = self.loop.create_future()
        notifications = G90DeviceNotifications(sock=self.socket_mock)
        await notifications.listen()
        self.socket_mock.recvfrom.side_effect = sock_data_awaitable
        asynctest.set_read_ready(self.socket_mock, self.loop)
        with self.assertLogs(level='WARNING') as cm:
            await asyncio.wait([future], timeout=0.1)
            self.assertEqual(cm.output, [
                'WARNING:pyg90alarm.device_notifications:'
                'Unknown notification received from mocked:12345: kind 999,'
                ' data [1]'
            ])
        notifications.close()

    async def test_unknown_device_alert(self):
        def sock_data_awaitable(*args):
            future.set_result(True)
            return (b'[208,[999,100,1,1,"Hall","DUMMYGUID",'
                    b'1631545189,0,[""]]]\0', ('mocked', 12345))
        future = self.loop.create_future()
        notifications = G90DeviceNotifications(sock=self.socket_mock)
        await notifications.listen()
        self.socket_mock.recvfrom.side_effect = sock_data_awaitable
        asynctest.set_read_ready(self.socket_mock, self.loop)
        with self.assertLogs(level='WARNING') as cm:
            await asyncio.wait([future], timeout=0.1)
            self.assertEqual(cm.output, [
                'WARNING:pyg90alarm.device_notifications:'
                'Unknown alert received from mocked:12345: type 999,'
                ' data G90DeviceAlert(type=999, event_id=100, source=1,'
                " state=1, zone_name='Hall', device_id='DUMMYGUID',"
                " unix_time=1631545189, resv4=0, other=[''])"
            ])
        notifications.close()

    async def test_sensor_callback(self):
        future = self.loop.create_future()
        sensor_cb = MagicMock()
        sensor_cb.side_effect = lambda *args: future.set_result(True)
        notifications = G90DeviceNotifications(
            sensor_cb=sensor_cb, sock=self.socket_mock)
        await notifications.listen()
        asynctest.set_read_ready(self.socket_mock, self.loop)
        self.socket_mock.recvfrom.return_value = (
            b'[170,[5,[100,"Hall"]]]\0', ('mocked', 12345))
        await asyncio.wait([future], timeout=0.1)
        notifications.close()
        sensor_cb.assert_called_once_with(100, 'Hall')

    async def test_armdisarm_notification_callback(self):
        future = self.loop.create_future()
        armdisarm_cb = MagicMock()
        armdisarm_cb.side_effect = lambda *args: future.set_result(True)
        notifications = G90DeviceNotifications(
            armdisarm_cb=armdisarm_cb, sock=self.socket_mock)
        await notifications.listen()
        asynctest.set_read_ready(self.socket_mock, self.loop)
        self.socket_mock.recvfrom.return_value = (
            b'[170,[1,[1]]]\0', ('mocked', 12345))
        await asyncio.wait([future], timeout=0.1)
        notifications.close()
        armdisarm_cb.assert_called_once_with(1)

    async def test_armdisarm_alert_callback(self):
        future = self.loop.create_future()
        armdisarm_cb = MagicMock()
        armdisarm_cb.side_effect = lambda *args: future.set_result(True)
        notifications = G90DeviceNotifications(
            armdisarm_cb=armdisarm_cb, sock=self.socket_mock)
        await notifications.listen()
        asynctest.set_read_ready(self.socket_mock, self.loop)
        self.socket_mock.recvfrom.return_value = (
            b'[208,[2,4,0,0,"","DUMMYGUID",1630876128,0,[""]]]\0',
            ('mocked', 12345))
        await asyncio.wait([future], timeout=0.1)
        notifications.close()
        armdisarm_cb.assert_called_once_with(1)

    async def test_door_open_close_callback(self):
        future = self.loop.create_future()
        door_open_close_cb = MagicMock()
        door_open_close_cb.side_effect = lambda *args: future.set_result(True)
        notifications = G90DeviceNotifications(
            door_open_close_cb=door_open_close_cb, sock=self.socket_mock)
        await notifications.listen()
        asynctest.set_read_ready(self.socket_mock, self.loop)
        self.socket_mock.recvfrom.return_value = (
            b'[208,[4,100,1,1,"Hall","DUMMYGUID",1631545189,0,[""]]]\0',
            ('mocked', 12345))
        await asyncio.wait([future], timeout=0.1)
        notifications.close()
        door_open_close_cb.assert_called_once_with(100, 'Hall', True)

    async def test_doorbell_callback(self):
        future = self.loop.create_future()
        door_open_close_cb = MagicMock()
        door_open_close_cb.side_effect = lambda *args: future.set_result(True)
        notifications = G90DeviceNotifications(
            door_open_close_cb=door_open_close_cb, sock=self.socket_mock)
        await notifications.listen()
        asynctest.set_read_ready(self.socket_mock, self.loop)
        self.socket_mock.recvfrom.return_value = (
            b'[208,[4,111,12,0,"Doorbell","DUMMYGUID",1655745021,0,[""]]]\0',
            ('mocked', 12345))
        await asyncio.wait([future], timeout=0.1)
        notifications.close()
        door_open_close_cb.assert_called_once_with(111, 'Doorbell', True)



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Protocol entity for G90 alarm panel information.
"""

from collections import namedtuple
from enum import IntEnum

INCOMING_FIELDS = [
    'host_guid',
    'product_name',
    'wifi_protocol_version',
    'cloud_protocol_version',
    'mcu_hw_version',
    'wifi_hw_version',
    'gsm_status_data',
    'wifi_status_data',
    'reserved1',
    'reserved2',
    'band_frequency',
    'gsm_signal_level',
    'wifi_signal_level'
]


class G90HostInfoGsmStatus(IntEnum):
    """
    Defines possible values of GSM module status.
    """
    POWERED_OFF = 0
    SIM_ABSENT = 1
    NO_SIGNAL = 2
    OPERATIONAL = 3


class G90HostInfoWifiStatus(IntEnum):
    """
    Defines possible values of Wifi module status.
    """
    POWERED_OFF = 0
    NOT_CONNECTED = 1
    NO_SIGNAL = 2
    OPERATIONAL = 3


class G90HostInfo(namedtuple('G90HostInfo', INCOMING_FIELDS)):
    """
    Interprets data fields of GETHOSTINFO command.
    """
    @property
    def gsm_status(self):
        """
        Translates the GSM module status received from the device into
        corresponding enum.

        :return: :class:`G90HostInfoGsmStatus`
        """
        return G90HostInfoGsmStatus(self.gsm_status_data)

    @property
    def wifi_status(self):
        """
        Translates the Wifi module status received from the device into
        corresponding enum.

        :return: :class:`G90HostInfoWifiStatus`
        """
        return G90HostInfoWifiStatus(self.wifi_status_data)
